/* ========================================================================== *
!*
!* SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
!* http://software.intel.com/en-us/articles/
!*        intel-sample-source-code-license-agreement/
!*
!* Copyright 2014 Intel Corporation
!*
!* THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED,
!* INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY,
!* FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL
!* PROPERTY RIGHTS.
!*
!* ========================================================================== */
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <offload.h>
#include <sys/time.h>
#include <new>

#define MYMIC 0

// This is currently required to disable allocation/de-allocation for the
// implicit offload model when -no-offload was used. A future compiler version
// will do this automatically. (DPD200509198)
#ifndef __INTEL_OFFLOAD
#define _Offload_shared_aligned_malloc _mm_malloc
#define _Offload_shared_aligned_free _mm_free
#endif

// Timer implementation
static struct timeval tstart, tstop;
void start_timer()
{
    gettimeofday(&tstart, NULL);
}

double stop_timer()
{
    gettimeofday(&tstop, NULL);
    return (tstop.tv_sec + tstop.tv_usec / 1000000.0) -
           (tstart.tv_sec + tstart.tv_usec / 1000000.0);
}

_Cilk_shared
int init()
{
    int numThreads = 0;
    kmp_set_defaults("KMP_AFFINITY=scatter");
#pragma omp parallel for
    for (size_t i = 0; i < 10; i++) {
        if (i == 0) numThreads = omp_get_num_threads();
    }

    return numThreads;
}

class _Cilk_shared Data {
public:
    Data(size_t iS, size_t kS) : inputSize(iS), kernelSize(kS)
    {
        input = (_Cilk_shared float *)_Offload_shared_aligned_malloc((inputSize + kernelSize - 1) * sizeof(float), 64);
        kernel = (_Cilk_shared float *)_Offload_shared_aligned_malloc(kernelSize * sizeof(float), 64);
        output = (_Cilk_shared float *)_Offload_shared_aligned_malloc(inputSize * sizeof(float), 64);
        output[0:inputSize] = 0.0;
    }

    ~Data()
    {
        _Offload_shared_aligned_free(input);
        _Offload_shared_aligned_free(kernel);
        _Offload_shared_aligned_free(output);
    }

    _Cilk_shared float *input;
    _Cilk_shared float *kernel;
    _Cilk_shared float *output;
    size_t inputSize;
    size_t kernelSize;
};

// 1D linear convolution - the actual work
_Cilk_shared
void convolve(_Cilk_shared Data *data)
{
    float *input = data->input;
    float *kernel = data->kernel;
    float *output = data->output;
    __assume_aligned(input, 64);
    __assume_aligned(kernel, 64);
    __assume_aligned(output, 64);

    size_t inputSize = data->inputSize;
    size_t kernelSize = data->kernelSize;

    for (int run = 0; run < 200; run++) {
#pragma omp parallel for
        for (size_t i = 0; i < inputSize; i++) {
#pragma ivdep
            output[i] += __sec_reduce_add(kernel[0:kernelSize] * input[i:kernelSize]);
        }

        data->kernel[0:data->kernelSize] /= 10.0;
    }
}

__attribute__((aligned(64))) _Cilk_shared Data * _Cilk_shared data;

int main(int argc, char **argv)
{
    int numThreads = 0;
    double duration = 0.0;
    FILE *fhInput = NULL;
    FILE *fhKernel = NULL;
    FILE *fhOutput = NULL;

    if (argc != 4) {
        printf("Error: Wrong use!\n");
        printf("Usage: %s <input> <kernel> <output>\n", argv[0]);
        return 1;
    }

    fhInput = fopen(argv[1], "rb");
    if (fhInput == NULL) {
        printf("Error: Cannot read input file!\n");
        return 1;
    }
    fhKernel = fopen(argv[2], "rb");
    if (fhKernel == NULL) {
        printf("Error: Cannot read kernel file!\n");
        return 1;
    }
    fhOutput = fopen(argv[3], "wb");
    if (fhOutput == NULL) {
        printf("Error: Cannot open output file!\n");
        return 1;
    }

    // Get sizes of input and kernel
    fseek(fhInput, 0, SEEK_END);
    size_t inputSize = ftell(fhInput) / sizeof(float);
    rewind(fhInput);

    if (inputSize == 0) {
        printf("Error: Input file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of input file: %d\n", inputSize);

    fseek(fhKernel, 0, SEEK_END);
    size_t kernelSize = ftell(fhKernel) / sizeof(float);
    rewind(fhKernel);

    if (kernelSize == 0) {
        printf("Error: Kernel file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of kernel file: %d\n", kernelSize);

    // Create the data structure.
    // Note that since it is shared it requires dedicated allocation for the
    // implicit offload model. We simply use the placement new operator.
    data = new (_Offload_shared_aligned_malloc(sizeof(Data), 64)) _Cilk_shared Data(inputSize, kernelSize);

    // Read in input and kernel
    for(int i = 0; i < kernelSize - 1; i++) data->input[i] = 0; // Fade in

    float *temp = (float *)malloc(inputSize * sizeof(float));
    size_t curElement = 0;
    while (curElement != inputSize) {
        curElement += fread(&temp[curElement], sizeof(float), inputSize - curElement, fhInput);
    }
    memcpy(&data->input[kernelSize - 1], temp, inputSize * sizeof(float));
    free(temp);

    temp = (float *)malloc(kernelSize * sizeof(float));
    curElement = 0;
    while (curElement != kernelSize) {
        curElement += fread(&temp[curElement], sizeof(float), kernelSize - curElement, fhKernel);
    }
    memcpy(&data->kernel[0], temp, kernelSize * sizeof(float));
    free(temp);

    // Initialize the OpenMP runtime once and inform about available threads
    numThreads = _Cilk_offload init();
    printf("Number of OpenMP threads used: %d\n", numThreads);

    start_timer();
    _Cilk_offload_to(MYMIC) convolve(data);
    duration = stop_timer();

    fwrite(data->output, sizeof(float), inputSize, fhOutput);

    // Print results
    printf("Elapsed time = %lf seconds\n", duration);

    // Clean up
    fclose(fhInput);
    fclose(fhKernel);
    fclose(fhOutput);
    data->~Data();
    _Offload_shared_aligned_free(data);
    return 0;
}
